@@ -1,7 +1,8 @@
 from ctypes import cdll, CDLL, RTLD_GLOBAL
-from ctypes import POINTER, byref, c_int, c_int64, c_int32, c_bool, c_char_p, c_double, c_void_p, CFUNCTYPE, py_object, cast, byref, Structure, string_at
+from ctypes import POINTER, byref, c_int, c_int64, c_int32, c_bool, c_char_p, c_double, c_void_p, CFUNCTYPE, py_object, cast, byref, Structure, string_at, pointer
 import ctypes
 import traceback
+from typing import Any
 '''
 CDLL(MPI.__file__, mode=RTLD_GLOBAL) is a workaround for a few MPICH bugs, including 
 the bug with non-working MPI_IN_PLACE and 2-stage ELPA solver
@@ -20,8 +21,42 @@ from pathlib import Path
 import os, shutil
 from ase import units
 from scalapack4py import ScaLAPACK4py
-from scalapack4py.blacsdesc import ctypes2ndarray
+from scalapack4py.array_types import ctypes2ndarray
 from scipy.linalg import lapack
+from scipy.linalg import logm, expm
+from scipy.spatial.transform import Rotation
+
+def spharm_rot(l_qn:int, R:Rotation) -> np.ndarray:
+  '''
+    Returns rotation matrix for spherical harmonics
+    
+    Parameters
+    ----------
+    l_qn : Angular momentum quantum number
+    R : Rotation descriptor
+
+    Returns
+    -------
+    DM : ndarray[2*l_qn+1, 2*l_qn+1] Rotation matrix
+  '''
+  if not hasattr(spharm_rot, 'Jx'):
+    from sphecerix import tesseral_wigner_D
+    
+    spharm_rot.Jx = [logm(tesseral_wigner_D(l, Rotation.from_rotvec([1,0,0]))) for l in range(12)]
+    spharm_rot.Jy = [logm(tesseral_wigner_D(l, Rotation.from_rotvec([0,1,0]))) for l in range(12)]
+    spharm_rot.Jz = [logm(tesseral_wigner_D(l, Rotation.from_rotvec([0,0,1]))) for l in range(12)]
+  
+  
+  ax,ay, az = R.as_euler('xyz')
+  Ax = expm(spharm_rot.Jx[l_qn] * ax)
+  Ay = expm(spharm_rot.Jy[l_qn] * ay)
+  Az = expm(spharm_rot.Jz[l_qn] * az)
+  A = Az @ Ay @ Ax
+  assert A.shape[0] == l_qn*2+1
+  
+  Condon_Shortley_m_sign = [1 if m <= 0 else (-1)**m for m in range(-l_qn, l_qn+1)] # that is for AIMS only
+
+  return np.einsum("i,ij,j->ij", Condon_Shortley_m_sign, A, Condon_Shortley_m_sign)
 
 class ASI_matrix_descr(Structure):
     _fields_ = [("matrix_type", c_int),
@@ -38,22 +73,39 @@ class ASI_basis_func_descr_generic(Structure):
      in basis_func_descr_ptr_types dict
   '''
   _fields_ = [("descr_size",c_int), ("basis_func_descr_type",c_int),]
+  
+  def copy(self):
+    dst = type(self)()
+    pointer(dst)[0] = self
+    return dst
 
 class ASI_basis_func_descr_generic_local(Structure):
   _fields_ = [("descr_size",c_int), ("basis_func_descr_type",c_int), ("atom_index",c_int),]
 
 class ASI_basis_func_descr_aims(Structure):
-    _fields_ = [("descr_size",c_int), ("basis_func_descr_type",c_int), \
-    ("atom_index",c_int),
-    ("func_type_ptr", c_char_p),
-    ("func_type_len", c_int),
-    ("n_qn",c_int),
-    ("l_qn",c_int),
-    ("m_qn",c_int)]
-    
-    @property
-    def func_type(self):
-      return string_at(self.func_type_ptr,self.func_type_len).decode('utf-8')
+  _fields_ = [("descr_size",c_int), ("basis_func_descr_type",c_int), \
+  ("atom_index",c_int),
+  ("func_type_ptr", c_char_p),
+  ("func_type_len", c_int),
+  ("n_qn",c_int),
+  ("l_qn",c_int),
+  ("m_qn",c_int)]
+
+  rot_invariant_attrs = ["atom_index", "func_type", "n_qn", "l_qn"]
+
+  @property
+  def func_type(self):
+    return string_at(self.func_type_ptr,self.func_type_len).decode('utf-8')
+
+  def copy(self):
+    print("copy:", self)
+    dst = type(self)()
+    pointer(dst)[0] = self
+    return dst
+  
+  def __repr__(self):
+    l = 'spdfghiklmnop'[self.l_qn]
+    return f'a{self.atom_index} {self.func_type} {self.n_qn}{l}{self.m_qn}'
 
 basis_func_descr_types = {101:ASI_basis_func_descr_generic_local, 102:ASI_basis_func_descr_aims}
 
@@ -159,7 +211,7 @@ def default_loading_callback(aux, iK, iS, descr, data, matrix_descr_ptr):
     m = storage_dict[(iK, iS)] if asi.scalapack.is_root(descr) else None
     assert m is None or (asi.n_basis == m.shape[0]) and (asi.n_basis == m.shape[1]), \
                      f"m.shape=={m.shape} != asi.n_basis=={asi.n_basis}"
-    asi.scalapack.scatter_numpy(m, descr, data)
+    asi.scalapack.scatter_numpy(m, descr, data, asi.hamiltonian_dtype)
     return 1 # signal that  matrix has been loaded
   except Exception as eee:
     print (f"Something happened in ASI default_loading_callback {label}: {eee}\nAborting...")
@@ -169,6 +221,7 @@ def default_loading_callback(aux, iK, iS, descr, data, matrix_descr_ptr):
         traceback = traceback.tb_next
     MPI.COMM_WORLD.Abort(1)
 
+
 class ASIlib:
   '''
     Python wrapper for dynamically loaded library with ASI API implementation
@@ -248,6 +301,10 @@ class ASIlib:
       self.lib.ASI_register_hamiltonian_callback.argtypes = [dmhs_callback, c_void_p]
       if hasattr(self.lib, "ASI_register_dm_init_callback"):
         self.lib.ASI_register_dm_init_callback.argtypes = [set_dmhs_callback, c_void_p]
+      if hasattr(self.lib, "ASI_register_set_hamiltonian_callback"):
+        self.lib.ASI_register_set_hamiltonian_callback.argtypes = [set_dmhs_callback, c_void_p]
+      if hasattr(self.lib, "ASI_register_modify_hamiltonian_callback"):
+        self.lib.ASI_register_modify_hamiltonian_callback.argtypes = [set_dmhs_callback, c_void_p]
       self.lib.ASI_register_external_potential.argtypes = [esp_callback, c_void_p];
       self.lib.ASI_is_hamiltonian_real.restype = c_bool
       self.lib.ASI_get_basis_size.restype = c_int
@@ -261,6 +318,8 @@ class ASIlib:
       self.lib.ASI_get_local_ks.argtypes = [ndpointer(dtype=np.int32),]
       if hasattr(self.lib, "ASI_is_scf_converged"):
         self.lib.ASI_is_scf_converged.restype = c_bool
+      if hasattr(self.lib, "ASI_get_k_points"):
+        self.lib.ASI_get_k_points.restype = POINTER(c_double)
       
       input_filename = {1:"dummy", 2:"dftb_in.hsd"}[self.lib.ASI_flavour()]
       self.lib.ASI_init(input_filename.encode('UTF-8'), self.logfile.encode('UTF-8'), c_int(self.mpi_comm.py2f()))
@@ -357,6 +416,42 @@ class ASIlib:
     self.hamiltonian_aux = hamiltonian_aux
     self.lib.ASI_register_hamiltonian_callback(self.hamiltonian_callback, c_void_p.from_buffer(py_object(self.hamiltonian_aux)))
 
+  def register_set_hamiltonian_callback(self, set_hamiltonian_callback, set_hamiltonian_aux):
+    """
+      Register callback function to be called when evaluating the hamitlonian
+      in the first SCF iteration
+      
+      Calls `ASI_register_set_hamiltonian_callback()`
+
+      Parameters
+      ----------
+      set_hamiltonian_callback : dmhs_callback
+        Callback function
+      set_hamiltonian_aux : Object
+        Auxiliary object for the callback
+    """
+    self.set_hamiltonian_callback = set_dmhs_callback(set_hamiltonian_callback)
+    self.set_hamiltonian_aux = set_hamiltonian_aux
+    self.lib.ASI_register_set_hamiltonian_callback(self.set_hamiltonian_callback, c_void_p.from_buffer(py_object(self.set_hamiltonian_aux)))
+
+  def register_modify_hamiltonian_callback(self, modify_hamiltonian_callback, modify_hamiltonian_aux):
+    """
+      Register callback function that adds a hamiltonian-shaped array to the hamiltonian calculated
+      in the DFT code.
+      
+      Calls `ASI_register_modify_hamiltonian_callback()`
+
+      Parameters
+      ----------
+      modify_hamiltonian_callback : dmhs_callback
+        Callback function
+      modify_hamiltonian_aux : Object
+        Auxiliary object for the callback
+    """
+    self.modify_hamiltonian_callback = set_dmhs_callback(modify_hamiltonian_callback)
+    self.modify_hamiltonian_aux = modify_hamiltonian_aux
+    self.lib.ASI_register_modify_hamiltonian_callback(self.modify_hamiltonian_callback, c_void_p.from_buffer(py_object(self.modify_hamiltonian_aux)))
+    
   def register_dm_callback(self, dm_callback, dm_aux):
     """
       Register callback function to be called on Density Matrix calculation
@@ -460,13 +555,82 @@ class ASIlib:
     '''
     return (self.n_basis,self.n_basis) if self.is_hamiltonian_real else (self.n_basis,self.n_basis, 2) # resulting shape
   
-  def get_basis_func_descr(self, bf_index):
+  def get_basis_func_descr(self, bf_index:int) -> ASI_basis_func_descr_generic:
+    '''
+      Invokes ASI_get_basis_func_descr and returns descriptor of the requested basis function
+
+      Parameters
+      ----------
+      bf_index : int
+        Basis function index, from range `[0, n_basis)`.
+      
+      Returns
+      -------
+        ASI_basis_func_descr_generic instance or its derivative (ASI_basis_func_descr_generic_local or ASI_basis_func_descr_aims)
+    '''
     if not hasattr(self.lib, "ASI_get_basis_func_descr"):
       raise NotImplementedError("ASI_get_basis_func_descr not implemented in {self.lib_file}")
     
     descr_ptr = self.lib.ASI_get_basis_func_descr(bf_index)
     descr_ptr_specific = cast(descr_ptr, POINTER(basis_func_descr_types[descr_ptr.contents.basis_func_descr_type]))
     return descr_ptr_specific.contents
+  
+  @property
+  def basis_funcs(self) -> list[ASI_basis_func_descr_generic]:
+    '''
+      Returns list of basis function descriptors
+      
+      Returns
+      -------
+        List of instances of ASI_basis_func_descr_generic or of its derivatives (ASI_basis_func_descr_generic_local or ASI_basis_func_descr_aims)
+    '''
+    return [self.get_basis_func_descr(i).copy() for i in range(self.n_basis)]
+  
+  def get_basis_funcs_rotgrp(self) -> dict[Any, list[tuple[int, ASI_basis_func_descr_generic]]]:
+    '''
+      Returns indexed list of basis function descriptors for the specified atom 
+      grouped by rotationally-invariant attributes (all but `m_qn`)
+
+      Parameters
+      ----------
+      atom_index : int
+        Atom index, from range `[0, n_atoms)`.
+      
+      Returns
+      -------
+        Dict of enumerated lists of basis functions grouped by rotationally invariant attributes (such as n_qn and l_qn).
+        Enumeration index corresponds to index of the basis function in the total `basis_funcs` list.
+        Basis functions in groups are sorted by the `m_qn` attribute.
+    '''
+    res = dict()
+    for i,bf in enumerate(self.basis_funcs):
+      rotinv_key = tuple(getattr(bf, attrname) for attrname in bf.rot_invariant_attrs)
+      if rotinv_key not in res:
+        res[rotinv_key] = []
+      res[rotinv_key].append((i,bf))
+    
+    for k in res.keys():
+      v = res[k]
+      v = sorted(v, key=lambda i_bf: i_bf[1].m_qn)
+    return res
+  
+  def get_rotation_matrix(self, R):
+    from scipy.linalg import block_diag
+    rotgrps = self.get_basis_funcs_rotgrp()
+    
+    for k,v in rotgrps.items():
+      print ("rotgrps", k, v)
+    
+    rot_matrices = []
+    bf_indices = []
+    for bf_list in rotgrps.values():
+      l_qn = bf_list[0][1].l_qn # l_qn should be the same for all basis functions in any group
+      assert len(bf_list) == 2*l_qn+1, f'{len(bf_list)} != {2*l_qn+1}'
+      rot_matrices.append(spharm_rot(l_qn, R))
+      bf_indices += [bf_index for bf_index,_ in bf_list]
+    bf_indices = np.array(bf_indices)
+    A = block_diag(*rot_matrices)[bf_indices, :][:, bf_indices]
+    return A
 
   @property
   def basis_atoms(self):
@@ -495,6 +659,17 @@ class ASIlib:
     """
     return self.lib.ASI_get_nkpts()
 
+  @property
+  def k_points(self):
+    """
+      Calls `ASI_get_k_points()`
+      Returns: ndarray of shape (n_kpts,3) with k-point coordinates in relative 
+      units of reciprocal  lattice vectors
+    """
+    shape = (self.n_kpts, 3)
+    kpts = np.ctypeslib.as_array(self.lib.ASI_get_k_points(), shape=shape)
+    return kpts
+
   @property
   def n_local_ks(self):
     """
@@ -515,7 +690,7 @@ class ASIlib:
     res = np.zeros((n*2,), dtype=c_int32)
     n2 =  self.lib.ASI_get_local_ks(res)
     assert n == n2
-    return res
+    return res.reshape((n, 2))
 
   @property
   def is_hamiltonian_real(self):
@@ -526,6 +701,15 @@ class ASIlib:
     """
     return self.lib.ASI_is_hamiltonian_real()
 
+  @property
+  def hamiltonian_dtype(self):
+    """
+      numpy.dtype: numpy.float64  if Hamiltonian of current system is real. numpy.complex128 if Hamiltonian of current system is complex.
+      
+      Calls `ASI_is_hamiltonian_real()`
+    """
+    return np.float64 if self.is_hamiltonian_real else np.complex128
+
   @property
   def is_scf_converged(self):
     """
@@ -674,3 +858,29 @@ class ASIlib:
     self.dm_init_storage = value
     self.register_DM_init(default_loading_callback, (self, self.dm_init_storage, 'DM init'))
 
+  @property
+  def set_hamiltonian(self):
+    """
+      bool / c_double[n_basis, n_basis] : Set with hamiltonian to be initialised at the first SCF step
+      Reading that property returns True if the hamiltonian setting is enabled
+    """
+    return hasattr(self, 'set_hamiltonian_callback')
+ 
+  @set_hamiltonian.setter
+  def set_hamiltonian(self, value):
+    self.set_H_storage = value
+    self.register_set_hamiltonian_callback(default_loading_callback, (self, self.set_H_storage, 'Set H'))
+
+  @property
+  def modify_hamiltonian(self):
+    """
+      bool / c_double[n_basis, n_basis] : Adds a hamiltonian-shaped array to the hamiltonian calculated
+                                          in the QM code. 
+      Reading that property returns True if the hamiltonian modify is enabled
+    """
+    return hasattr(self, 'modify_hamiltonian_callback')
+ 
+  @modify_hamiltonian.setter
+  def modify_hamiltonian(self, value):
+    self.modify_H_storage = value
+    self.register_modify_hamiltonian_callback(default_loading_callback, (self, self.modify_H_storage, 'Modify H'))
